@@ -1,7 +1,7 @@
 
 /*  $Header$
  *
- *  Copyright (c) 1987-2007 Sun Microsystems, Inc. All Rights Reserved.
+ *  Copyright (c) 1987-2008 Sun Microsystems, Inc. All Rights Reserved.
  *           
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -143,11 +143,11 @@ static struct button_widget button_widgets[] = {
     { GDK_f, 0 }},
 
     {KEY_CLEAR,              "clear_simple",
-    { 0, 0 },
+    { GDK_SHIFT_MASK, 0 },
     { GDK_Delete, 0 }},
     
     {KEY_CLEAR,              "clear_advanced",
-    { 0, 0 },
+    { GDK_SHIFT_MASK, 0 },
     { GDK_Delete, 0 }},
 
     {KEY_SHIFT,              "shift_left",
@@ -197,7 +197,7 @@ static struct button_widget button_widgets[] = {
     {KEY_BACKSPACE,          "backspace_advanced",
     { 0, 0 },
     { GDK_BackSpace, 0 }},
-
+    
     {KEY_NUMERIC_POINT,      "numeric_point",
     { 0,          0,              0,             0 },
     { GDK_period, GDK_KP_Decimal, GDK_KP_Delete, GDK_KP_Separator, 0 }},
@@ -223,8 +223,8 @@ static struct button_widget button_widgets[] = {
     { GDK_minus, GDK_KP_Subtract, GDK_R4, 0 }},
 
     {KEY_MULTIPLY,           "multiply",
-    { GDK_SHIFT_MASK, 0,               0,     0,      0 },
-    { GDK_asterisk,   GDK_KP_Multiply, GDK_x, GDK_R6, 0 }},
+    { GDK_SHIFT_MASK, 0,               0,      0 },
+    { GDK_asterisk,   GDK_KP_Multiply, GDK_R6, 0 }},
 
     {KEY_DIVIDE,             "divide",
     { 0,         GDK_SHIFT_MASK, 0,             0,      GDK_SHIFT_MASK, 0 },
@@ -302,6 +302,10 @@ static struct button_widget button_widgets[] = {
     { GDK_SHIFT_MASK, 0 },
     { GDK_G, 0 }},
 
+    {KEY_LOGARITHM2,         "logarithm2",
+    { GDK_SHIFT_MASK, 0 },
+    { GDK_H, 0 }},
+
     {KEY_FACTORIAL,          "factorial",
     { GDK_SHIFT_MASK, 0 },
     { GDK_exclam, 0 }},
@@ -335,8 +339,8 @@ static struct button_widget button_widgets[] = {
     { GDK_ampersand, 0 }},
 
     {KEY_XOR,                "xor",
-    { 0 },
-    { 0 }},
+    { 0,     0 },
+    { GDK_x, 0 }},
 
     {KEY_XNOR,               "xnor",
     { 0, 0 },
@@ -387,10 +391,6 @@ static struct button_widget button_widgets[] = {
 #define GET_WIDGET(name) \
           glade_xml_get_widget(X->ui, (name))
 
-#define SET_MENUBAR_ITEM_STATE(name, state) \
-          g_object_set_data(G_OBJECT(GET_WIDGET(name)), "sensitive", \
-                            GINT_TO_POINTER(state));
-
 #define CONNECT_SIGNAL(name) glade_xml_signal_connect(X->ui, #name, \
                        G_CALLBACK(name))
 
@@ -488,7 +488,7 @@ reset_display(void)
         case EXPRS:
             e = get_state();
             MPstr_to_num("0", DEC, e->ans);
-            exp_del();
+            exp_clear();
             show_display(e->ans);
             break;
         
@@ -763,11 +763,22 @@ ui_set_mode(enum mode_type mode)
     char *hostname, title[MAXLINE];
     GtkWidget *menu;
 
-    refresh_display(-1);
-    ui_make_registers();
+    if (v->modetype != mode) {
+        v->modetype = mode;
+
+        ui_set_base(DEC);
+        ui_set_numeric_mode(FIX);
+        ui_set_accuracy(DEFAULT_ACCURACY);
+        ui_set_show_thousands_seperator(FALSE);
+        ui_set_show_trailing_zeroes(FALSE);
+        ui_make_registers();
+
+        /* Reset display */
+        do_clear();
+    }
     
     /* Save mode */
-    set_resource(R_MODE, Rmstr[(int) v->modetype]);
+    set_resource(R_MODE, Rmstr[(int)mode]);
     
     /* Show/enable the widgets used in this mode */
     g_object_set(G_OBJECT(X->bas_panel),  "visible", mode == BASIC, NULL);
@@ -866,58 +877,18 @@ ui_set_statusbar(gchar *text, const gchar *imagename)
     gtk_statusbar_push(GTK_STATUSBAR(X->statusbar), 0, text); 
 }
 
-
-static void
-bin_str(int MP_value[MP_SIZE], char *str, int maxbits)
-{
-    int i, MP0[MP_SIZE], MP1[MP_SIZE], MP2[MP_SIZE], MP3[MP_SIZE];
-    int neg = 0;
-
-    MPstr_to_num("0", DEC, MP0);
-    MPstr_to_num("1", DEC, MP1);
-    MPstr_to_num("2", DEC, MP2);
-
-    if (mplt(MP_value, MP0)) {
-        mpcmim(MP_value, MP0);
-        mpadd(MP0, MP1, MP0);
-        neg = 1;
-    } else {
-        mpcmim(MP_value, MP0);
-    }
-    
-    for (i = 0; i < maxbits; i++) {
-        double lsb;               /* Least significant bit. */
-        calc_and(MP3, MP0, MP1);
-        mpcmd(MP3, &lsb);
-
-        if (lsb == 0) {
-            str[maxbits - i -1] = (neg) ? '1' : '0';
-        } else {
-            str[maxbits - i -1] = (neg) ? '0' : '1';
-        } 
-
-        mpdiv(MP0, MP2, MP3);
-        mpcmim(MP3, MP0);
-    }
-
-    return;
-}
-
-
 static void
 set_bit_panel()
 {
     int bit_str_len, i, MP1[MP_SIZE], MP2[MP_SIZE];
     int MP[MP_SIZE];
-    char str[64], label[3];
-    int ret = usable_num(MP);
+    char *bit_str, label[3], tmp[MAXLINE];
 
     switch (v->syntax) {
         case NPA:
             MPstr_to_num(v->display, v->base, MP1);
             mpcmim(MP1, MP2);
             if (mpeq(MP1, MP2)) {
-                char *bit_str, label[3], tmp[MAXLINE];
                 int toclear = (v->current == KEY_CLEAR_ENTRY)
                               ? TRUE : FALSE;
 
@@ -938,19 +909,29 @@ set_bit_panel()
             gtk_widget_set_sensitive(X->bit_panel, FALSE);
             break;
 
-        case EXPRS:
-            if (ret || !is_integer(MP)) {
-                gtk_widget_set_sensitive(X->bit_panel, FALSE);
-                return;
-            }
-            bin_str(MP, str, 64);
-            gtk_widget_set_sensitive(X->bit_panel, TRUE);
-
-            STRCPY(label, " 0");
-            for (i = 0; i < 64; i++) {
-                label[1] = str[64 - i - 1];
-                gtk_label_set_text(GTK_LABEL(X->bits[64 - i - 1]), label);
-            }
+          case EXPRS: 
+              {
+                  int ret = usable_num(MP);
+                  if (ret || !is_integer(MP)) {
+                      gtk_widget_set_sensitive(X->bit_panel, FALSE);
+                      return;
+                  }
+                  char *bit_str, label[3], tmp[MAXLINE];
+                  bit_str = make_fixed(MP, tmp, BIN, MAXLINE, FALSE);
+                  bit_str_len = strlen(bit_str);
+                  if (bit_str_len <= MAXBITS) {
+                      gtk_widget_set_sensitive(X->bit_panel, TRUE);
+                      
+                      STRCPY(label, " 0");
+                      for (i = 0; i < MAXBITS; i++) {
+                          label[1] = (i < bit_str_len) ? bit_str[bit_str_len-i-1] : '0';
+                          gtk_label_set_text(GTK_LABEL(X->bits[MAXBITS - i - 1]), label);
+                      }
+                  } else {
+                      gtk_widget_set_sensitive(X->bit_panel, FALSE);
+                  }
+                  
+              }
             break;
         
         default:
@@ -1037,20 +1018,28 @@ ui_set_error_state(gboolean error)
 
     gtk_widget_set_sensitive(X->mode_panel, !v->error);
 
-    // FIXME: Isn't this missing a whole lot of widgets?
-    SET_MENUBAR_ITEM_STATE("copy_menu",            !v->error);
-    SET_MENUBAR_ITEM_STATE("paste_menu",           !v->error); 
-    SET_MENUBAR_ITEM_STATE("insert_ascii_menu",    !v->error); 
-    SET_MENUBAR_ITEM_STATE("view_basic_menu",      !v->error); 
-    SET_MENUBAR_ITEM_STATE("view_advanced_menu",   !v->error); 
-    SET_MENUBAR_ITEM_STATE("view_financial_menu",  !v->error); 
-    SET_MENUBAR_ITEM_STATE("view_scientific_menu", !v->error); 
-    SET_MENUBAR_ITEM_STATE("show_trailing_zeroes_menu",
-                           !v->error && (v->modetype == SCIENTIFIC)); 
-    SET_MENUBAR_ITEM_STATE("show_thousands_separator_menu",  !v->error); 
-    SET_MENUBAR_ITEM_STATE("show_registers_menu",  !v->error); 
-    SET_MENUBAR_ITEM_STATE("arithmetic_precedence_menu", !v->error); 
-    SET_MENUBAR_ITEM_STATE("about_menu",           !v->error);
+    gtk_widget_set_sensitive(GET_WIDGET("copy_menu"),            !v->error);
+    gtk_widget_set_sensitive(GET_WIDGET("paste_menu"),           !v->error); 
+    gtk_widget_set_sensitive(GET_WIDGET("undo_menu"),            !v->error);
+    gtk_widget_set_sensitive(GET_WIDGET("redo_menu"),            !v->error);
+    gtk_widget_set_sensitive(GET_WIDGET("insert_ascii_menu"),    !v->error); 
+
+    gtk_widget_set_sensitive(GET_WIDGET("view_basic_menu"),      !v->error); 
+    gtk_widget_set_sensitive(GET_WIDGET("view_advanced_menu"),   !v->error); 
+    gtk_widget_set_sensitive(GET_WIDGET("view_financial_menu"),  !v->error); 
+    gtk_widget_set_sensitive(GET_WIDGET("view_scientific_menu"), !v->error); 
+    gtk_widget_set_sensitive(GET_WIDGET("show_trailing_zeroes_menu"),
+                             !v->error && (v->modetype == SCIENTIFIC)); 
+    gtk_widget_set_sensitive(GET_WIDGET("show_thousands_separator_menu"),
+                             !v->error); 
+    gtk_widget_set_sensitive(GET_WIDGET("show_bitcalculating_menu"), !v->error);
+    gtk_widget_set_sensitive(GET_WIDGET("show_registers_menu"), !v->error); 
+    gtk_widget_set_sensitive(GET_WIDGET("ltr_precedence_menu"), !v->error);
+
+    gtk_widget_set_sensitive(GET_WIDGET("arithmetic_precedence_menu"),
+                             !v->error); 
+
+    gtk_widget_set_sensitive(GET_WIDGET("about_menu"), !v->error);
 }
 
 
@@ -1178,7 +1167,7 @@ about_cb(GtkWidget *widget)
     gtk_show_about_dialog(GTK_WINDOW(X->kframe),
             "name",_("Gcalctool"),
             "version", VERSION,
-            "copyright", _("\xc2\xa9 1986-2007 The Gcalctool authors"),
+            "copyright", _("\xc2\xa9 1986-2008 The Gcalctool authors"),
             "license", license_trans,
             "comments", _("Calculator with financial and scientific modes."),
             "authors", authors,
@@ -1659,23 +1648,6 @@ save_win_position()
 }
 
 
-static void
-change_mode(int mode)
-{
-    v->modetype = mode;
-
-    ui_set_base(DEC);
-    ui_set_numeric_mode(FIX);
-    ui_set_accuracy(DEFAULT_ACCURACY);
-    ui_set_show_thousands_seperator(FALSE);
-    ui_set_show_trailing_zeroes(FALSE);
-    ui_set_mode(v->modetype);
-
-    /* Reset display */
-    do_clear();
-}
-
-
 static gboolean
 request_change_mode()
 {
@@ -1724,7 +1696,7 @@ request_change_mode()
     // FIXME: Save this in GConf
     X->warn_change_mode = !gtk_toggle_button_get_active(
                              GTK_TOGGLE_BUTTON(request_check));
-    
+
     gtk_widget_destroy(dialog);
 
     return (response == GTK_RESPONSE_ACCEPT);
@@ -2022,13 +1994,14 @@ kframe_key_press_cb(GtkWidget *widget, GdkEventKey *event)
     int i, j, state;
     GtkWidget *button;
     
+    /* Only look at the modifiers we use */
+    state = event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK);
+    
     if (check_for_localized_numeric_point(event->keyval) == TRUE) {
         event->state = 0;
         event->keyval = GDK_KP_Decimal;
     }
     
-    state = event->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK);
-
     /* Accuracy shortcuts */
     if (state == GDK_CONTROL_MASK && v->modetype == SCIENTIFIC) {
         switch (event->keyval) {
@@ -2075,6 +2048,12 @@ kframe_key_press_cb(GtkWidget *widget, GdkEventKey *event)
             return (TRUE);
         }
     }
+    
+    /* Delete in display */
+    if (event->keyval == GDK_Delete && state == 0) {
+        do_button(KEY_DELETE, 0);
+        return (TRUE);
+    }
 
     for (i = 0; i < NBUTTONS; i++) {
         button = X->buttons[i];
@@ -2085,14 +2064,12 @@ kframe_key_press_cb(GtkWidget *widget, GdkEventKey *event)
             continue;
         }
 
-        j = 0;
-        while (button_widgets[i].accelerator_keys[j] != 0) {
+        for (j = 0; button_widgets[i].accelerator_keys[j] != 0; j++) {
             if (button_widgets[i].accelerator_keys[j] == event->keyval &&
-                (button_widgets[i].accelerator_mods[j] & ~GDK_SHIFT_MASK) == state) {
+                button_widgets[i].accelerator_mods[j] == state) {
                 button_cb(button, NULL);
                 return (TRUE);
             }
-            j++;
         }
     }
 
@@ -2374,8 +2351,10 @@ mode_radio_cb(GtkWidget *menu)
         v->modetype = mode;
         ui_set_mode(v->modetype);
     } else {
-        if (request_change_mode()) {
-            change_mode(mode);
+        if (mode != v->modetype && request_change_mode()) {
+            ui_set_mode(mode);
+        } else {
+            ui_set_mode(v->modetype);
         }
     }
 }
@@ -2546,7 +2525,7 @@ create_kframe()
                                         GTK_BUTTONS_NONE,
                                         N_("Error loading user interface"));
         gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
-                                                 N_("The user interace file %s is missing or unable to be loaded. Please check your installation."), UI_FILE);
+                                                 N_("The user interface file %s is missing or unable to be loaded. Please check your installation."), UI_FILE);
         gtk_dialog_add_buttons(GTK_DIALOG(dialog), GTK_STOCK_QUIT, GTK_RESPONSE_ACCEPT, NULL);
         
         gtk_dialog_run(GTK_DIALOG(dialog));
@@ -2661,7 +2640,7 @@ create_kframe()
     for (i = 0; i < NBUTTONS; i++) {
         SNPRINTF(name, MAXLINE, "calc_%s_button", 
                  button_widgets[i].widget_name);
-        X->buttons[i] = GET_WIDGET(name);
+        X->buttons[i] = GET_WIDGET(name);            
         assert(X->buttons[i] != NULL);
         
         gtk_size_group_add_widget(size_group, X->buttons[i]);
